@@ -47,6 +47,19 @@ void report_path_error_ex(const char *ps, json_path_t *p,
 void report_json_error_ex(const char *js, json_engine_t *je,
                           const char *fname, int n_param,
                           Sql_condition::enum_warning_level lv);
+int check_overlaps(json_engine_t *js, json_engine_t *value, bool compare_whole);
+int json_find_overlap_with_object(json_engine_t *js,
+                                              json_engine_t *value,
+                                              bool compare_whole);
+void json_skip_current_level(json_engine_t *js, json_engine_t *value);
+bool json_find_overlap_with_scalar(json_engine_t *js, json_engine_t *value);
+bool json_compare_arrays_in_order_in_order(json_engine_t *js, json_engine_t *value);
+bool json_compare_arr_and_obj(json_engine_t *js, json_engine_t* value);
+int json_find_overlap_with_array(json_engine_t *js,
+                                             json_engine_t *value,
+                                             bool compare_whole);
+
+
 
 class Json_engine_scan: public json_engine_t
 {
@@ -88,9 +101,9 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_valid") };
     return name;
   }
-  bool fix_length_and_dec() override
+  bool fix_length_and_dec(THD *thd) override
   {
-    if (Item_bool_func::fix_length_and_dec())
+    if (Item_bool_func::fix_length_and_dec(thd))
       return TRUE;
     set_maybe_null();
     return FALSE;
@@ -107,6 +120,23 @@ public:
 };
 
 
+class Item_func_json_equals: public Item_bool_func
+{
+public:
+  Item_func_json_equals(THD *thd, Item *a, Item *b):
+    Item_bool_func(thd, a, b) {}
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_equals") };
+    return name;
+  }
+  bool fix_length_and_dec(THD *thd) override;
+  Item *do_get_copy(THD *thd) const override
+  { return get_item_copy<Item_func_json_equals>(thd, this); }
+  longlong val_int() override;
+};
+
+
 class Item_func_json_exists: public Item_bool_func
 {
 protected:
@@ -121,7 +151,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_exists") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_exists>(thd, this); }
   longlong val_int() override;
@@ -158,7 +188,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_value") };
     return name;
   }
-  bool fix_length_and_dec() override ;
+  bool fix_length_and_dec(THD *thd) override ;
   String *val_str(String *to) override
   {
     null_value= Json_path_extractor::extract(to, args[0], args[1],
@@ -186,7 +216,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_query") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *to) override
   {
     null_value= Json_path_extractor::extract(to, args[0], args[1],
@@ -215,7 +245,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_quote") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_quote>(thd, this); }
@@ -234,7 +264,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_unquote") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_unquote>(thd, this); }
@@ -285,7 +315,7 @@ public:
     return name;
   }
   enum Functype functype() const override { return JSON_EXTRACT_FUNC; }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   longlong val_int() override;
   double val_real() override;
@@ -312,7 +342,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_contains") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   longlong val_int() override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_contains>(thd, this); }
@@ -339,7 +369,7 @@ public:
     return name;
   }
   bool fix_fields(THD *thd, Item **ref) override;
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   longlong val_int() override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_contains_path>(thd, this); }
@@ -357,7 +387,7 @@ public:
   Item_func_json_array(THD *thd, List<Item> &list):
     Item_json_func(thd, list) {}
   String *val_str(String *) override;
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   LEX_CSTRING func_name_cstring() const override
   {
     static LEX_CSTRING name= {STRING_WITH_LEN("json_array") };
@@ -376,7 +406,7 @@ protected:
 public:
   Item_func_json_array_append(THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list) {}
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   uint get_n_paths() const override { return arg_count/2; }
   LEX_CSTRING func_name_cstring() const override
@@ -455,6 +485,24 @@ public:
   { return get_item_copy<Item_func_json_merge_patch>(thd, this); }
 };
 
+
+class Item_func_json_normalize: public Item_json_func
+{
+public:
+  Item_func_json_normalize(THD *thd, Item *a):
+    Item_json_func(thd, a) {}
+  String *val_str(String *) override;
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_normalize") };
+    return name;
+  }
+  bool fix_length_and_dec(THD *thd) override;
+  Item *do_get_copy(THD *thd) const override
+  { return get_item_copy<Item_func_json_normalize>(thd, this); }
+};
+
+
 class Item_func_json_length: public Item_long_func
 {
   bool check_arguments() const override
@@ -480,7 +528,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_length") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   longlong val_int() override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_length>(thd, this); }
@@ -500,7 +548,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_depth") };
     return name;
   }
-  bool fix_length_and_dec() override { max_length= 10; return FALSE; }
+  bool fix_length_and_dec(THD *thd) override { max_length= 10; return FALSE; }
   longlong val_int() override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_depth>(thd, this); }
@@ -518,7 +566,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_type") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_type>(thd, this); }
@@ -535,7 +583,7 @@ public:
   Item_func_json_insert(bool i_mode, bool r_mode, THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list),
       mode_insert(i_mode), mode_replace(r_mode) {}
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   uint get_n_paths() const override { return arg_count/2; }
   LEX_CSTRING func_name_cstring() const override
@@ -558,7 +606,7 @@ protected:
 public:
   Item_func_json_remove(THD *thd, List<Item> &list):
     Item_json_str_multipath(thd, list) {}
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   uint get_n_paths() const override { return arg_count - 1; }
   LEX_CSTRING func_name_cstring() const override
@@ -585,7 +633,7 @@ public:
     static LEX_CSTRING name= {STRING_WITH_LEN("json_keys") };
     return name;
   }
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   Item *do_get_copy(THD *thd) const override
   { return get_item_copy<Item_func_json_keys>(thd, this); }
@@ -613,7 +661,7 @@ public:
     return name;
   }
   bool fix_fields(THD *thd, Item **ref) override;
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *) override;
   uint get_n_paths() const override { return arg_count > 4 ? arg_count - 4 : 0; }
   Item *do_get_copy(THD *thd) const override
@@ -641,7 +689,7 @@ public:
     Item_json_func(thd, list), fmt(DETAILED) {}
 
   LEX_CSTRING func_name_cstring() const override;
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   String *val_str(String *str) override;
   String *val_json(String *str) override;
   Item *do_get_copy(THD *thd) const override
@@ -745,4 +793,23 @@ public:
 
 extern bool is_json_type(const Item *item);
 
+class Item_func_json_overlaps: public Item_bool_func
+{
+  String tmp_js;
+  bool a2_constant, a2_parsed;
+  String tmp_val, *val;
+public:
+  Item_func_json_overlaps(THD *thd, Item *a, Item *b):
+    Item_bool_func(thd, a, b) {}
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("json_overlaps") };
+    return name;
+  }
+  bool fix_length_and_dec(THD *thd) override;
+  longlong val_int() override;
+  Item *do_get_copy(THD *thd) const override
+  { return get_item_copy<Item_func_json_overlaps>(thd, this); }
+};
+
 #endif /* ITEM_JSONFUNC_INCLUDED */
