@@ -87,7 +87,7 @@ protected:
   { }
 
 protected:
-  virtual void change_env(THD *thd) const
+  void change_env(THD *thd) const override
   {
     thd->variables.collation_database= m_db_cl;
 
@@ -519,6 +519,18 @@ private:
                                                sp_assignment_lex *param_lex,
                                                Item_args *parameters);
 
+  bool bind_input_param(THD *thd,
+                        Item *arg_item,
+                        uint arg_no,
+                        sp_rcontext *nctx,
+                        bool is_function);
+
+  bool bind_output_param(THD *thd,
+                         Item *arg_item,
+                         uint arg_no,
+                         sp_rcontext *octx,
+                         sp_rcontext *nctx);
+
 public:
   /**
     Generate a code for an "OPEN cursor" statement.
@@ -575,7 +587,7 @@ public:
   { return m_flags & MODIFIES_DATA; }
 
   inline uint instructions()
-  { return m_instr.elements; }
+  { return (uint)m_instr.elements; }
 
   inline sp_instr *
   last_instruction()
@@ -1044,9 +1056,9 @@ public:
     return m_routine_implementations.check_dup_qualified(lex->sphead) ||
            m_routine_implementations.push_back(lex, &main_mem_root);
   }
-  sp_package *get_package() { return this; }
-  void init_psi_share();
-  bool is_invoked() const
+  sp_package *get_package() override { return this; }
+  void init_psi_share() override;
+  bool is_invoked() const override
   {
     /*
       Cannot flush a package out of the SP cache when:
@@ -1081,7 +1093,7 @@ public:
     Query_arena(thd->lex->sphead->get_main_mem_root(), STMT_INITIALIZED_FOR_SP)
   { }
   ~sp_lex_cursor() { free_items(); }
-  virtual bool cleanup_stmt(bool /*restore_set_statement_vars*/) override
+  bool cleanup_stmt(bool /*restore_set_statement_vars*/) override
   { return false; }
   Query_arena *query_arena() override { return this; }
   bool validate()
@@ -1126,7 +1138,7 @@ public:
   sp_instr(uint ip, sp_pcontext *ctx)
     :Query_arena(0, STMT_INITIALIZED_FOR_SP), marked(0), m_ip(ip), m_ctx(ctx)
 #ifdef PROTECT_STATEMENT_MEMROOT
-  , m_has_been_run(false)
+  , m_has_been_run(NON_RUN)
 #endif
   {}
 
@@ -1221,21 +1233,29 @@ public:
 #ifdef PROTECT_STATEMENT_MEMROOT
   bool has_been_run() const
   {
-    return m_has_been_run;
+    return m_has_been_run == RUN;
+  }
+
+  void mark_as_qc_used()
+  {
+    m_has_been_run= QC;
   }
 
   void mark_as_run()
   {
-    m_has_been_run= true;
+    if (m_has_been_run == QC)
+      m_has_been_run= NON_RUN; // answer was from WC => not really executed
+    else
+      m_has_been_run= RUN;
   }
 
   void mark_as_not_run()
   {
-    m_has_been_run= false;
+    m_has_been_run= NON_RUN;
   }
 
 private:
-  bool m_has_been_run;
+  enum {NON_RUN, QC, RUN} m_has_been_run;
 #endif
 }; // class sp_instr : public Sql_alloc
 
@@ -1351,18 +1371,18 @@ public:
 
   virtual ~sp_instr_stmt() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 private:
 
   sp_lex_keeper m_lex_keeper;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 
 }; // class sp_instr_stmt : public sp_instr
@@ -1386,11 +1406,11 @@ public:
 
   virtual ~sp_instr_set() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 protected:
   sp_rcontext *get_rcontext(THD *thd) const;
@@ -1400,7 +1420,7 @@ protected:
   sp_lex_keeper m_lex_keeper;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_set : public sp_instr
 
@@ -1429,9 +1449,9 @@ public:
 
   virtual ~sp_instr_set_row_field() = default;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 }; // class sp_instr_set_field : public sp_instr_set
 
 
@@ -1471,9 +1491,9 @@ public:
 
   virtual ~sp_instr_set_row_field_by_name() = default;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 }; // class sp_instr_set_field_by_name : public sp_instr_set
 
 
@@ -1497,11 +1517,11 @@ public:
 
   virtual ~sp_instr_set_trigger_field() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 private:
   Item_trigger_field *trigger_field;
@@ -1509,7 +1529,7 @@ private:
   sp_lex_keeper m_lex_keeper;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_trigger_field : public sp_instr
 
@@ -1543,7 +1563,7 @@ public:
   virtual void set_destination(uint old_dest, uint new_dest)
     = 0;
 
-  virtual uint get_cont_dest() const;
+  uint get_cont_dest() const override;
 
 protected:
 
@@ -1569,17 +1589,17 @@ public:
 
   virtual ~sp_instr_jump() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads);
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override;
 
-  virtual uint opt_shortcut_jump(sp_head *sp, sp_instr *start);
+  uint opt_shortcut_jump(sp_head *sp, sp_instr *start) override;
 
-  virtual void opt_move(uint dst, List<sp_instr> *ibp);
+  void opt_move(uint dst, List<sp_instr> *ibp) override;
 
-  virtual void backpatch(uint dest, sp_pcontext *dst_ctx)
+  void backpatch(uint dest, sp_pcontext *dst_ctx) override
   {
     /* Calling backpatch twice is a logic flaw in jump resolution. */
     DBUG_ASSERT(m_dest == 0);
@@ -1589,14 +1609,14 @@ public:
   /**
     Update the destination; used by the optimizer.
   */
-  virtual void set_destination(uint old_dest, uint new_dest)
+  void set_destination(uint old_dest, uint new_dest) override
   {
     if (m_dest == old_dest)
       m_dest= new_dest;
   }
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_jump : public sp_instr_opt_meta
 
@@ -1620,23 +1640,23 @@ public:
 
   virtual ~sp_instr_jump_if_not() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads);
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override;
 
   /** Override sp_instr_jump's shortcut; we stop here */
-  virtual uint opt_shortcut_jump(sp_head *sp, sp_instr *start)
+  uint opt_shortcut_jump(sp_head *sp, sp_instr *start) override
   {
     return m_ip;
   }
 
-  virtual void opt_move(uint dst, List<sp_instr> *ibp);
+  void opt_move(uint dst, List<sp_instr> *ibp) override;
 
-  virtual void set_destination(uint old_dest, uint new_dest)
+  void set_destination(uint old_dest, uint new_dest) override
   {
     sp_instr_jump::set_destination(old_dest, new_dest);
     if (m_cont_dest == old_dest)
@@ -1649,7 +1669,7 @@ private:
   sp_lex_keeper m_lex_keeper;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_jump_if_not : public sp_instr_jump
 
@@ -1667,18 +1687,18 @@ public:
 
   virtual ~sp_instr_preturn() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads)
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override
   {
     marked= 1;
     return UINT_MAX;
   }
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_preturn : public sp_instr
 
@@ -1698,13 +1718,13 @@ public:
 
   virtual ~sp_instr_freturn() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads)
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override
   {
     marked= 1;
     return UINT_MAX;
@@ -1717,7 +1737,7 @@ protected:
   sp_lex_keeper m_lex_keeper;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_freturn : public sp_instr
 
@@ -1746,19 +1766,19 @@ public:
     m_handler= NULL;
   }
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads);
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override;
 
   /** Override sp_instr_jump's shortcut; we stop here. */
-  virtual uint opt_shortcut_jump(sp_head *sp, sp_instr *start)
+  uint opt_shortcut_jump(sp_head *sp, sp_instr *start) override
   {
     return m_ip;
   }
 
-  virtual void backpatch(uint dest, sp_pcontext *dst_ctx)
+  void backpatch(uint dest, sp_pcontext *dst_ctx) override
   {
     DBUG_ASSERT(!m_dest || !m_opt_hpop);
     if (!m_dest)
@@ -1785,7 +1805,7 @@ private:
   uint m_frame;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_hpush_jump : public sp_instr_jump
 
@@ -1808,16 +1828,16 @@ public:
     m_count= count;
   }
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 private:
 
   uint m_count;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_hpop : public sp_instr
 
@@ -1836,24 +1856,24 @@ public:
 
   virtual ~sp_instr_hreturn() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
   /* This instruction will not be short cut optimized. */
-  virtual uint opt_shortcut_jump(sp_head *sp, sp_instr *start)
+  uint opt_shortcut_jump(sp_head *sp, sp_instr *start) override
   {
     return m_ip;
   }
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads);
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override;
 
 private:
 
   uint m_frame;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_hreturn : public sp_instr_jump
 
@@ -1881,7 +1901,7 @@ public:
     cursor is closed. For now stored procedures always use materialized
     cursors and the call is not used.
   */
-  virtual bool cleanup_stmt(bool /*restore_set_statement_vars*/) override
+  bool cleanup_stmt(bool /*restore_set_statement_vars*/) override
   { return false; }
 private:
 
@@ -1912,16 +1932,16 @@ public:
     m_count= count;
   }
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 private:
 
   uint m_count;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_cpop : public sp_instr
 
@@ -1939,18 +1959,18 @@ public:
 
   virtual ~sp_instr_copen() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 private:
 
   uint m_cursor;		///< Stack index
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_copen : public sp_instr_stmt
 
@@ -1975,12 +1995,12 @@ public:
       m_var(voffs)
   {}
   virtual ~sp_instr_cursor_copy_struct() = default;
-  virtual int execute(THD *thd, uint *nextp);
-  virtual int exec_core(THD *thd, uint *nextp);
-  virtual void print(String *str);
+  int execute(THD *thd, uint *nextp) override;
+  int exec_core(THD *thd, uint *nextp) override;
+  void print(String *str) override;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 };
 
@@ -1998,16 +2018,16 @@ public:
 
   virtual ~sp_instr_cclose() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 private:
 
   uint m_cursor;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_cclose : public sp_instr
 
@@ -2027,9 +2047,9 @@ public:
 
   virtual ~sp_instr_cfetch() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
   void add_to_varlist(sp_variable *var)
   {
@@ -2043,7 +2063,7 @@ private:
   bool m_error_on_no_data;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_cfetch : public sp_instr
 
@@ -2065,12 +2085,12 @@ public:
 
   virtual ~sp_instr_agg_cfetch() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_agg_cfetch : public sp_instr
 
@@ -2090,11 +2110,11 @@ public:
 
   virtual ~sp_instr_error() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads)
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override
   {
     marked= 1;
     return UINT_MAX;
@@ -2105,7 +2125,7 @@ private:
   int m_errcode;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_error : public sp_instr
 
@@ -2123,17 +2143,17 @@ public:
 
   virtual ~sp_instr_set_case_expr() = default;
 
-  virtual int execute(THD *thd, uint *nextp);
+  int execute(THD *thd, uint *nextp) override;
 
-  virtual int exec_core(THD *thd, uint *nextp);
+  int exec_core(THD *thd, uint *nextp) override;
 
-  virtual void print(String *str);
+  void print(String *str) override;
 
-  virtual uint opt_mark(sp_head *sp, List<sp_instr> *leads);
+  uint opt_mark(sp_head *sp, List<sp_instr> *leads) override;
 
-  virtual void opt_move(uint dst, List<sp_instr> *ibp);
+  void opt_move(uint dst, List<sp_instr> *ibp) override;
 
-  virtual void set_destination(uint old_dest, uint new_dest)
+  void set_destination(uint old_dest, uint new_dest) override
   {
     if (m_cont_dest == old_dest)
       m_cont_dest= new_dest;
@@ -2146,7 +2166,7 @@ private:
   sp_lex_keeper m_lex_keeper;
 
 public:
-  virtual PSI_statement_info* get_psi_info() { return & psi_info; }
+  PSI_statement_info* get_psi_info() override { return & psi_info; }
   static PSI_statement_info psi_info;
 }; // class sp_instr_set_case_expr : public sp_instr_opt_meta
 
