@@ -82,7 +82,7 @@ public:
     :Stored_program_creation_ctx(client_cs, connection_cl, db_cl)
   { }
 
-  virtual Stored_program_creation_ctx *clone(MEM_ROOT *mem_root)
+  Stored_program_creation_ctx *clone(MEM_ROOT *mem_root) override
   {
     return new (mem_root) Trigger_creation_ctx(m_client_cs,
                                                m_connection_cl,
@@ -90,7 +90,7 @@ public:
   }
 
 protected:
-  virtual Object_creation_ctx *create_backup_ctx(THD *thd) const
+  Object_creation_ctx *create_backup_ctx(THD *thd) const override
   {
     return new Trigger_creation_ctx(thd);
   }
@@ -282,8 +282,8 @@ public:
   Handle_old_incorrect_sql_modes_hook(const char *file_path)
     :path(file_path)
   {};
-  virtual bool process_unknown_string(const char *&unknown_key, uchar* base,
-                                      MEM_ROOT *mem_root, const char *end);
+  bool process_unknown_string(const char *&unknown_key, uchar* base,
+                                      MEM_ROOT *mem_root, const char *end) override;
 };
 
 
@@ -294,8 +294,8 @@ public:
                                           LEX_CSTRING *trigger_table_arg)
     :path(file_path), trigger_table_value(trigger_table_arg)
   {};
-  virtual bool process_unknown_string(const char *&unknown_key, uchar* base,
-                                      MEM_ROOT *mem_root, const char *end);
+  bool process_unknown_string(const char *&unknown_key, uchar* base,
+                                      MEM_ROOT *mem_root, const char *end) override;
 private:
   const char *path;
   LEX_CSTRING *trigger_table_value;
@@ -322,12 +322,12 @@ public:
 
   Deprecated_trigger_syntax_handler() : m_trigger_name(NULL) {}
 
-  virtual bool handle_condition(THD *thd,
+  bool handle_condition(THD *thd,
                                 uint sql_errno,
                                 const char* sqlstate,
                                 Sql_condition::enum_warning_level *level,
                                 const char* message,
-                                Sql_condition ** cond_hdl)
+                                Sql_condition ** cond_hdl) override
   {
     if (sql_errno != EE_OUTOFMEMORY &&
         sql_errno != ER_OUT_OF_RESOURCES)
@@ -470,7 +470,7 @@ bool mysql_create_or_drop_trigger(THD *thd, TABLE_LIST *tables, bool create)
   /*
     We don't allow creating triggers on tables in the 'mysql' schema
   */
-  if (create && lex_string_eq(&tables->db, STRING_WITH_LEN("mysql")))
+  if (create && Lex_ident_db(tables->db).streq(MYSQL_SCHEMA_NAME))
   {
     my_error(ER_NO_TRIGGERS_ON_SYSTEM_SCHEMA, MYF(0));
     DBUG_RETURN(TRUE);
@@ -983,7 +983,7 @@ bool Table_triggers_list::create_trigger(THD *thd, TABLE_LIST *tables,
   /* Use the filesystem to enforce trigger namespace constraints. */
   trigger_exists= !access(trigname_file.str, F_OK);
 
-  ddl_log_create_trigger(thd, ddl_log_state, &tables->db, &tables->table_name,
+  ddl_log_create_trigger(ddl_log_state, &tables->db, &tables->table_name,
                          &lex->spname->m_name,
                          trigger_exists || table->triggers->count ?
                          DDL_CREATE_TRIGGER_PHASE_DELETE_COPY :
@@ -992,7 +992,7 @@ bool Table_triggers_list::create_trigger(THD *thd, TABLE_LIST *tables,
   /* Make a backup of the .TRG file that we can restore in case of crash */
   if (table->triggers->count &&
       (sql_backup_definition_file(&file, &backup_name) ||
-       ddl_log_delete_tmp_file(thd, ddl_log_state_tmp_file, &backup_name,
+       ddl_log_delete_tmp_file(ddl_log_state_tmp_file, &backup_name,
                                ddl_log_state)))
     DBUG_RETURN(true);
 
@@ -1004,7 +1004,7 @@ bool Table_triggers_list::create_trigger(THD *thd, TABLE_LIST *tables,
 
       /* Make a backup of the .TRN file that we can restore in case of crash */
       if (sql_backup_definition_file(&trigname_file, &backup_name) ||
-          ddl_log_delete_tmp_file(thd, ddl_log_state_tmp_file, &backup_name,
+          ddl_log_delete_tmp_file(ddl_log_state_tmp_file, &backup_name,
                                   ddl_log_state))
         DBUG_RETURN(true);
       ddl_log_update_phase(ddl_log_state, DDL_CREATE_TRIGGER_PHASE_OLD_COPIED);
@@ -1350,7 +1350,7 @@ bool Table_triggers_list::drop_trigger(THD *thd, TABLE_LIST *tables,
       /* This code is executed in case of DROP TRIGGER */
       lex_string_set3(&query, thd->query(), thd->query_length());
     }
-    if (ddl_log_drop_trigger(thd, ddl_log_state,
+    if (ddl_log_drop_trigger(ddl_log_state,
                              &tables->db, &tables->table_name,
                              sp_name, &query))
       goto err;
