@@ -648,7 +648,8 @@ public:
       */
       DBUG_ASSERT(item->type_handler()->type_handler_base_or_self()->
                   is_traditional_scalar_type() ||
-                  item->type_handler() == type_handler());
+                  item->type_handler()->type_collection() ==
+                  type_handler()->type_collection());
       return Data_type_compatibility::OK;
     }
     /**
@@ -663,7 +664,8 @@ public:
       // See the DBUG_ASSERT comment in can_optimize_keypart_ref()
       DBUG_ASSERT(item->type_handler()->type_handler_base_or_self()->
                   is_traditional_scalar_type() ||
-                  item->type_handler() == type_handler());
+                  item->type_handler()->type_collection() ==
+                  type_handler()->type_collection());
       return Data_type_compatibility::OK;
     }
     void hash_not_null(Hasher *hasher) override
@@ -935,7 +937,7 @@ public:
       str->append(singleton()->name().lex_cstring());
       str->append(')');
     }
-    bool fix_length_and_dec() override
+    bool fix_length_and_dec(THD *thd) override
     {
       Type_std_attributes::operator=(Type_std_attributes_fbt());
       if (Fbt::fix_fields_maybe_null_on_conversion_to_fbt(args[0]))
@@ -1126,6 +1128,11 @@ public:
     return FbtImpl::max_char_length();
   }
 
+  const Type_handler *type_handler_for_implicit_upgrade() const override
+  {
+    return TypeCollectionImpl::singleton()->
+             type_handler_for_implicit_upgrade(this);
+  }
   const Type_handler *type_handler_for_comparison() const override
   {
     return this;
@@ -1209,10 +1216,10 @@ public:
     return false;
   }
 
-  bool Column_definition_prepare_stage1(THD *thd, MEM_ROOT *mem_root,
+  bool Column_definition_prepare_stage1(THD *, MEM_ROOT *,
                                         Column_definition *def,
-                                        handler *file, ulonglong table_flags,
-                                        const Column_derived_attributes *derived_attr)
+                                        column_definition_type_t,
+                                        const Column_derived_attributes *)
                                         const override
   {
     def->prepare_stage1_simple(&my_charset_numeric);
@@ -1944,6 +1951,12 @@ public:
     return NULL;
   }
 
+  const Type_handler *type_handler_for_implicit_upgrade(
+                                               const Type_handler *from) const
+  {
+    return from;
+  }
+
   static Type_collection_fbt *singleton()
   {
     static Type_collection_fbt tc;
