@@ -24,13 +24,13 @@ Modified           Jan Lindström jan.lindstrom@mariadb.com
 *******************************************************/
 
 #include "fil0crypt.h"
-#include "mtr0types.h"
 #include "mach0data.h"
 #include "page0zip.h"
 #include "buf0checksum.h"
 #ifdef UNIV_INNOCHECKSUM
 # include "buf0buf.h"
 #else
+#include "buf0flu.h"
 #include "buf0dblwr.h"
 #include "srv0srv.h"
 #include "srv0start.h"
@@ -1348,6 +1348,8 @@ inline bool fil_space_t::acquire_if_not_stopped()
 
 bool fil_crypt_must_default_encrypt()
 {
+  /* prevents a race condition with fil_crypt_set_rotate_key_age() */
+  mysql_mutex_assert_owner(&fil_system.mutex);
   return !srv_fil_crypt_rotate_key_age || !srv_encrypt_rotate;
 }
 
@@ -1678,8 +1680,9 @@ fil_crypt_get_page_throttle(
 		return NULL;
 	}
 
-	if (DB_SUCCESS_LOCKED_REC
-	    != fseg_page_is_allocated(space, state->offset)) {
+	if (offset % (zip_size ? zip_size : srv_page_size)
+	    && DB_SUCCESS_LOCKED_REC
+	    != fseg_page_is_allocated(space, offset)) {
 		/* page is already freed */
 		return NULL;
 	}
@@ -2200,6 +2203,27 @@ void fil_crypt_set_rotation_iops(uint val)
   mysql_mutex_unlock(&fil_crypt_threads_mutex);
 }
 
+/** Add the import tablespace to default_encrypt list
+if necessary and signal fil_crypt_threads
+@param space imported tablespace */
+void fil_crypt_add_imported_space(fil_space_t *space)
+{
+  mysql_mutex_lock(&fil_crypt_threads_mutex);
+
+  mysql_mutex_lock(&fil_system.mutex);
+
+  if (fil_crypt_must_default_encrypt())
+  {
+    fil_system.default_encrypt_tables.push_back(*space);
+    space->is_in_default_encrypt= true;
+  }
+
+  mysql_mutex_unlock(&fil_system.mutex);
+
+  pthread_cond_broadcast(&fil_crypt_threads_cond);
+  mysql_mutex_unlock(&fil_crypt_threads_mutex);
+}
+
 /*********************************************************************
 Adjust encrypt tables
 @param[in]	val		New setting for innodb-encrypt-tables */
@@ -2289,7 +2313,7 @@ void fil_space_crypt_close_tablespace(const fil_space_t *space)
 				   << space->chain.start->name << " ("
 				   << space->id << ") active threads "
 				   << crypt_data->rotate_state.active_threads
-				   << "flushing="
+				   << " flushing="
 				   << crypt_data->rotate_state.flushing << ".";
 			last = now;
 		}
