@@ -132,6 +132,7 @@ bool log_t::create()
 #endif
 
   latch.SRW_LOCK_INIT(log_latch_key);
+  lsn_lock.init();
 
   last_checkpoint_lsn= FIRST_LSN;
   log_capacity= 0;
@@ -236,9 +237,6 @@ void log_t::attach_low(log_file_t file, os_offset_t size)
       mprotect(ptr, size_t(size), PROT_READ);
       buf= static_cast<byte*>(ptr);
       max_buf_free= 1;
-# if defined __linux__ || defined _WIN32
-      set_block_size(CPU_LEVEL1_DCACHE_LINESIZE);
-# endif
       log_maybe_unbuffered= true;
       log_buffered= false;
       mtr_t::finisher_update();
@@ -273,13 +271,16 @@ void log_t::attach_low(log_file_t file, os_offset_t size)
                         log_buffered
                         ? "Buffered log writes"
                         : "File system buffers for log disabled",
-                        block_size);
+                        write_size);
 #endif
 
   mtr_t::finisher_update();
 #ifdef HAVE_PMEM
-  checkpoint_buf= static_cast<byte*>(aligned_malloc(block_size, block_size));
-  memset_aligned<64>(checkpoint_buf, 0, block_size);
+  ut_ad(ut_is_2pow(write_size));
+  ut_ad(write_size >= 512);
+  ut_ad(write_size <= 4096);
+  checkpoint_buf= static_cast<byte*>(aligned_malloc(write_size, write_size));
+  memset_aligned<512>(checkpoint_buf, 0, write_size);
   return true;
 #endif
 }
@@ -430,7 +431,7 @@ void log_t::set_buffered(bool buffered)
                           log_buffered
                           ? "Buffered log writes"
                           : "File system buffers for log disabled",
-                          block_size);
+                          write_size);
   }
   log_resize_release();
 }
@@ -467,6 +468,8 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
                           OS_FILE_NORMAL, OS_LOG_FILE, false, &success);
     if (success)
     {
+      ut_ad(!(size_t(file_size) & (write_size - 1)));
+      ut_ad(!(size_t(size) & (write_size - 1)));
       log_resize_release();
 
       void *ptr= nullptr, *ptr2= nullptr;
@@ -522,7 +525,7 @@ log_t::resize_start_status log_t::resize_start(os_offset_t size) noexcept
         {
           memcpy_aligned<16>(resize_buf, buf, (buf_free + 15) & ~15);
           start_lsn= first_lsn +
-            (~lsn_t{get_block_size() - 1} & (write_lsn - first_lsn));
+            (~lsn_t{write_size - 1} & (write_lsn - first_lsn));
         }
       }
       resize_lsn.store(start_lsn, std::memory_order_relaxed);
@@ -578,32 +581,30 @@ void log_t::resize_abort() noexcept
 
 /** Write an aligned buffer to ib_logfile0.
 @param buf    buffer to be written
-@param len    length of data to be written
+@param length length of data to be written
 @param offset log file offset */
-static void log_write_buf(const byte *buf, size_t len, lsn_t offset)
+static void log_write_buf(const byte *buf, size_t length, lsn_t offset)
 {
   ut_ad(write_lock.is_owner());
   ut_ad(!recv_no_log_write);
-  ut_d(const size_t block_size_1= log_sys.get_block_size() - 1);
+  ut_d(const size_t block_size_1= log_sys.write_size - 1);
   ut_ad(!(offset & block_size_1));
-  ut_ad(!(len & block_size_1));
+  ut_ad(!(length & block_size_1));
   ut_ad(!(size_t(buf) & block_size_1));
-  ut_ad(len);
+  ut_ad(length);
 
-  if (UNIV_LIKELY(offset + len <= log_sys.file_size))
+  const lsn_t maximum_write_length{log_sys.file_size - offset};
+  ut_ad(maximum_write_length <= log_sys.file_size - log_sys.START_OFFSET);
+
+  if (UNIV_UNLIKELY(length > maximum_write_length))
   {
-write:
-    log_sys.log.write(offset, {buf, len});
-    return;
+    log_sys.log.write(offset, {buf, size_t(maximum_write_length)});
+    length-= size_t(maximum_write_length);
+    buf+= size_t(maximum_write_length);
+    ut_ad(log_sys.START_OFFSET + length < offset);
+    offset= log_sys.START_OFFSET;
   }
-
-  const size_t write_len= size_t(log_sys.file_size - offset);
-  log_sys.log.write(offset, {buf, write_len});
-  len-= write_len;
-  buf+= write_len;
-  ut_ad(log_sys.START_OFFSET + len < offset);
-  offset= log_sys.START_OFFSET;
-  goto write;
+  log_sys.log.write(offset, {buf, length});
 }
 
 /** Invoke commit_checkpoint_notify_ha() to notify that outstanding
@@ -778,11 +779,12 @@ inline void log_t::persist(lsn_t lsn) noexcept
 }
 #endif
 
+ATTRIBUTE_COLD ATTRIBUTE_NOINLINE
 /** Write resize_buf to resize_log.
 @param length  the used length of resize_buf */
-ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
+void log_t::resize_write_buf(size_t length) noexcept
 {
-  const size_t block_size_1= get_block_size() - 1;
+  const size_t block_size_1= write_size - 1;
   ut_ad(!(resize_target & block_size_1));
   ut_ad(!(length & block_size_1));
   ut_ad(length > block_size_1);
@@ -802,7 +804,7 @@ ATTRIBUTE_COLD void log_t::resize_write_buf(size_t length) noexcept
   }
 
   ut_a(os_file_write_func(IORequestWrite, "ib_logfile101", resize_log.m_file,
-                          resize_flush_buf, offset, length) == DB_SUCCESS);
+                          buf, offset, length) == DB_SUCCESS);
 }
 
 /** Write buf to ib_logfile0.
@@ -824,64 +826,88 @@ template<bool release_latch> inline lsn_t log_t::write_buf() noexcept
   }
   else
   {
+    ut_ad(write_lock.is_owner());
     ut_ad(!recv_no_log_write);
     write_lock.set_pending(lsn);
     ut_ad(write_lsn >= get_flushed_lsn());
-    const size_t block_size_1{get_block_size() - 1};
-    lsn_t offset{calc_lsn_offset(write_lsn) & ~lsn_t{block_size_1}};
+    const size_t write_size_1{write_size - 1};
+    ut_ad(ut_is_2pow(write_size));
+    size_t length{buf_free.load(std::memory_order_relaxed)};
+    lsn_t offset{calc_lsn_offset(write_lsn)};
+    ut_ad(length >= (offset & write_size_1));
+    ut_ad(write_size_1 >= 511);
 
-    DBUG_PRINT("ib_log", ("write " LSN_PF " to " LSN_PF " at " LSN_PF,
-                          write_lsn, lsn, offset));
-    const byte *write_buf{buf};
-    size_t length{buf_free};
-    ut_ad(length >= (calc_lsn_offset(write_lsn) & block_size_1));
-    const size_t new_buf_free{length & block_size_1};
-    buf_free= new_buf_free;
-    ut_ad(new_buf_free == ((lsn - first_lsn) & block_size_1));
+    const byte *const write_buf{buf};
+    offset&= ~lsn_t{write_size_1};
 
-    if (new_buf_free)
+    if (length <= write_size_1)
     {
+      ut_ad(!((length ^ (size_t(lsn) - size_t(first_lsn))) & write_size_1));
+      /* Keep filling the same buffer until we have more than one block. */
 #if 0 /* TODO: Pad the last log block with dummy records. */
-      buf_free= log_pad(lsn, get_block_size() - new_buf_free,
-                        buf + new_buf_free, flush_buf);
+      buf_free= log_pad(lsn, (write_size_1 + 1) - length,
+                        buf + length, flush_buf);
       ... /* TODO: Update the LSN and adjust other code. */
 #else
-      /* The rest of the block will be written as garbage.
-      (We want to avoid memset() while holding exclusive log_sys.latch)
-      This block will be overwritten later, once records beyond
-      the current LSN are generated. */
 # ifdef HAVE_valgrind
-      MEM_MAKE_DEFINED(buf + length, get_block_size() - new_buf_free);
-      if (UNIV_LIKELY_NULL(resize_flush_buf))
-        MEM_MAKE_DEFINED(resize_buf + length, get_block_size() - new_buf_free);
+      MEM_MAKE_DEFINED(buf + length, (write_size_1 + 1) - length);
+      if (UNIV_LIKELY_NULL(resize_buf))
+        MEM_MAKE_DEFINED(resize_buf + length, (write_size_1 + 1) - length);
 # endif
       buf[length]= 0; /* allow recovery to catch EOF faster */
-      length&= ~block_size_1;
-      memcpy_aligned<16>(flush_buf, buf + length, (new_buf_free + 15) & ~15);
-      if (UNIV_LIKELY_NULL(resize_flush_buf))
-        memcpy_aligned<16>(resize_flush_buf, resize_buf + length,
-                           (new_buf_free + 15) & ~15);
-      length+= get_block_size();
 #endif
+      length= write_size_1 + 1;
+    }
+    else
+    {
+      const size_t new_buf_free{length & write_size_1};
+      ut_ad(new_buf_free == ((lsn - first_lsn) & write_size_1));
+      buf_free.store(new_buf_free, std::memory_order_relaxed);
+
+      if (new_buf_free)
+      {
+        /* The rest of the block will be written as garbage.
+        (We want to avoid memset() while holding exclusive log_sys.latch)
+        This block will be overwritten later, once records beyond
+        the current LSN are generated. */
+#ifdef HAVE_valgrind
+          MEM_MAKE_DEFINED(buf + length, (write_size_1 + 1) - new_buf_free);
+        if (UNIV_LIKELY_NULL(resize_buf))
+          MEM_MAKE_DEFINED(resize_buf + length, (write_size_1 + 1) -
+                           new_buf_free);
+#endif
+        buf[length]= 0; /* allow recovery to catch EOF faster */
+        length&= ~write_size_1;
+        memcpy_aligned<16>(flush_buf, buf + length, (new_buf_free + 15) & ~15);
+        if (UNIV_LIKELY_NULL(resize_buf))
+          memcpy_aligned<16>(resize_flush_buf, resize_buf + length,
+                             (new_buf_free + 15) & ~15);
+        length+= write_size_1 + 1;
+      }
+
+      std::swap(buf, flush_buf);
+      std::swap(resize_buf, resize_flush_buf);
     }
 
-    std::swap(buf, flush_buf);
-    std::swap(resize_buf, resize_flush_buf);
     write_to_log++;
     if (release_latch)
       latch.wr_unlock();
 
+    DBUG_PRINT("ib_log", ("write " LSN_PF " to " LSN_PF " at " LSN_PF,
+                          write_lsn, lsn, offset));
+
+    /* Do the write to the log file */
+    log_write_buf(write_buf, length, offset);
+
+    if (UNIV_LIKELY_NULL(resize_buf))
+      resize_write_buf(length);
+    write_lsn= lsn;
+
     if (UNIV_UNLIKELY(srv_shutdown_state > SRV_SHUTDOWN_INITIATED))
     {
       service_manager_extend_timeout(INNODB_EXTEND_TIMEOUT_INTERVAL,
                                      "InnoDB log write: " LSN_PF, write_lsn);
     }
-
-    /* Do the write to the log file */
-    log_write_buf(write_buf, length, offset);
-    if (UNIV_LIKELY_NULL(resize_buf))
-      resize_write_buf(length);
-    write_lsn= lsn;
   }
 
   set_check_for_checkpoint(false);
@@ -1305,6 +1331,7 @@ void log_t::close()
 #endif
 
   latch.destroy();
+  lsn_lock.destroy();
 
   recv_sys.close();
 
