@@ -59,16 +59,16 @@ public:
   const char         *mem_calc_func_name;
   const char         *mem_calc_file_name;
   ulong              mem_calc_line_no;
-  uint               sql_kinds;
-  uint               *sql_kind;
   ulonglong          *connection_ids;
   uint               conn_kinds;
   uint               *conn_kind;
   char               *conn_keys_first_ptr;
   char               **conn_keys;
   SPIDER_CONN        **conns;
-  /* for active-standby mode */
+  /* array of indexes of active servers */
   uint               *conn_link_idx;
+  /* A bitmap indicating whether each active server have some higher
+  numbered server in the same "group" left to try (can fail over) */
   uchar              *conn_can_fo;
   void               **quick_targets;
   int                *need_mons;
@@ -79,27 +79,11 @@ public:
   spider_string      *blob_buff;
   SPIDER_POSITION    *pushed_pos;
   SPIDER_POSITION    pushed_pos_buf;
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   SPIDER_PARTITION_HANDLER *partition_handler;
-#endif
+  /* Whether this ha_spider is the owner of its wide_handler. */
   bool                wide_handler_owner = FALSE;
   SPIDER_WIDE_HANDLER *wide_handler = NULL;
 
-#ifdef HA_CAN_BULK_ACCESS
-  int                pre_direct_init_result;
-  bool               is_bulk_access_clone;
-  bool               synced_from_clone_source;
-  bool               bulk_access_started;
-  bool               bulk_access_executing;
-  bool               bulk_access_pre_called;
-  SPIDER_BULK_ACCESS_LINK *bulk_access_link_first;
-  SPIDER_BULK_ACCESS_LINK *bulk_access_link_current;
-  SPIDER_BULK_ACCESS_LINK *bulk_access_link_exec_tgt;
-/*
-  bool               init_ha_mem_root;
-  MEM_ROOT           ha_mem_root;
-*/
-#endif
   bool               is_clone;
   ha_spider          *pt_clone_source_handler;
   ha_spider          *pt_clone_last_searcher;
@@ -111,31 +95,21 @@ public:
   bool               da_status;
   bool               use_spatial_index;
 
-#ifdef SPIDER_HAS_GROUP_BY_HANDLER
   uint                  idx_for_direct_join;
   bool                  use_fields;
   spider_fields         *fields;
   SPIDER_LINK_IDX_CHAIN *link_idx_chain;
   SPIDER_LINK_IDX_CHAIN *result_link_idx_chain;
-#endif
 
   /* for mrr */
   bool               mrr_with_cnt;
   uint               multi_range_cnt;
   uint               multi_range_hit_point;
-#ifdef HA_MRR_USE_DEFAULT_IMPL
   int                multi_range_num;
   bool               have_second_range;
   KEY_MULTI_RANGE    mrr_second_range;
   spider_string      *mrr_key_buff;
-#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100000
   range_id_t         *multi_range_keys;
-#else
-  char               **multi_range_keys;
-#endif
-#else
-  KEY_MULTI_RANGE    *multi_range_ranges;
-#endif
 
   char               *append_tblnm_alias;
   uint               append_tblnm_alias_length;
@@ -148,13 +122,11 @@ public:
   bool               use_pre_action;
   bool               pre_bitmap_checked;
   bool               bulk_insert;
-#ifdef HANDLER_HAS_NEED_INFO_FOR_AUTO_INC
   bool               info_auto_called;
-#endif
 #ifdef HANDLER_HAS_CAN_USE_FOR_AUTO_INC_INIT
   bool               auto_inc_temporary;
 #endif
-  int                bulk_size;
+  int                bulk_size= 0;
   int                direct_dup_insert;
   int                store_error_num;
   uint               dup_key_idx;
@@ -167,24 +139,15 @@ public:
 
   ulonglong          *db_request_id;
   uchar              *db_request_phase;
-  uchar              *m_handler_opened;
-  uint               *m_handler_id;
-  char               **m_handler_cid;
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
   bool               do_direct_update;
   uint               direct_update_kinds;
-#endif
   spider_index_rnd_init prev_index_rnd_init;
-#ifdef HANDLER_HAS_DIRECT_AGGREGATE
   SPIDER_ITEM_HLD    *direct_aggregate_item_first;
   SPIDER_ITEM_HLD    *direct_aggregate_item_current;
-#endif
   ha_rows            table_rows;
-#ifdef HA_HAS_CHECKSUM_EXTENDED
   ha_checksum        checksum_val;
   bool               checksum_null;
   uint               action_flags;
-#endif
 
   /* for fulltext search */
   bool               ft_init_and_first;
@@ -211,14 +174,14 @@ public:
   handler *clone(
     const char *name,
     MEM_ROOT *mem_root
-  );
+  ) override;
   const char **bas_ext() const;
   int open(
     const char* name,
     int mode,
     uint test_if_locked
-  );
-  int close();
+  ) override;
+  int close() override;
   int check_access_kind_for_connection(
     THD *thd,
     bool write_request
@@ -226,82 +189,62 @@ public:
   void check_access_kind(
     THD *thd
   );
-#ifdef HA_CAN_BULK_ACCESS
-  int additional_lock(
-    THD *thd,
-    enum thr_lock_type lock_type
-  );
-#endif
   THR_LOCK_DATA **store_lock(
     THD *thd,
     THR_LOCK_DATA **to,
     enum thr_lock_type lock_type
-  );
+  ) override;
   int external_lock(
     THD *thd,
     int lock_type
-  );
+  ) override;
   int start_stmt(
     THD *thd,
     thr_lock_type lock_type
-  );
-  int reset();
+  ) override;
+  int reset() override;
   int extra(
     enum ha_extra_function operation
-  );
-  int index_init(
-    uint idx,
-    bool sorted
-  );
-#ifdef HA_CAN_BULK_ACCESS
-  int pre_index_init(
-    uint idx,
-    bool sorted
-  );
-#endif
-  int index_end();
-#ifdef HA_CAN_BULK_ACCESS
-  int pre_index_end();
-#endif
+  ) override;
+  int index_init(uint idx, bool sorted) override;
+  int index_end() override;
   int index_read_map(
     uchar *buf,
     const uchar *key,
     key_part_map keypart_map,
     enum ha_rkey_function find_flag
-  );
+  ) override;
   int index_read_last_map(
     uchar *buf,
     const uchar *key,
     key_part_map keypart_map
-  );
+  ) override;
   int index_next(
     uchar *buf
-  );
+  ) override;
   int index_prev(
     uchar *buf
-  );
+  ) override;
   int index_first(
     uchar *buf
-  );
+  ) override;
   int index_last(
     uchar *buf
-  );
+  ) override;
   int index_next_same(
     uchar *buf,
     const uchar *key,
     uint keylen
-  );
+  ) override;
   int read_range_first(
     const key_range *start_key,
     const key_range *end_key,
     bool eq_range,
     bool sorted
-  );
-  int read_range_next();
+  ) override;
+  int read_range_next() override;
   void reset_no_where_cond();
   bool check_no_where_cond();
-#ifdef HA_MRR_USE_DEFAULT_IMPL
-#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100000
   ha_rows multi_range_read_info_const(
     uint keyno,
     RANGE_SEQ_IF *seq,
@@ -310,7 +253,7 @@ public:
     uint *bufsz,
     uint *flags,
     Cost_estimate *cost
-  );
+  ) override;
   ha_rows multi_range_read_info(
     uint keyno,
     uint n_ranges,
@@ -319,177 +262,100 @@ public:
     uint *bufsz,
     uint *flags,
     Cost_estimate *cost
-  );
-#else
-  ha_rows multi_range_read_info_const(
-    uint keyno,
-    RANGE_SEQ_IF *seq,
-    void *seq_init_param,
-    uint n_ranges,
-    uint *bufsz,
-    uint *flags,
-    COST_VECT *cost
-  );
-  ha_rows multi_range_read_info(
-    uint keyno,
-    uint n_ranges,
-    uint keys,
-    uint key_parts,
-    uint *bufsz,
-    uint *flags,
-    COST_VECT *cost
-  );
-#endif
+  ) override;
   int multi_range_read_init(
     RANGE_SEQ_IF *seq,
     void *seq_init_param,
     uint n_ranges,
     uint mode,
     HANDLER_BUFFER *buf
-  );
-#if defined(MARIADB_BASE_VERSION) && MYSQL_VERSION_ID >= 100000
+  ) override;
   int multi_range_read_next(
     range_id_t *range_info
-  );
+  ) override;
   int multi_range_read_next_first(
     range_id_t *range_info
   );
   int multi_range_read_next_next(
     range_id_t *range_info
   );
-#else
-  int multi_range_read_next(
-    char **range_info
-  );
-  int multi_range_read_next_first(
-    char **range_info
-  );
-  int multi_range_read_next_next(
-    char **range_info
-  );
-#endif
-#else
-  int read_multi_range_first(
-    KEY_MULTI_RANGE **found_range_p,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    HANDLER_BUFFER *buffer
-  );
-  int read_multi_range_next(
-    KEY_MULTI_RANGE **found_range_p
-  );
-#endif
-  int rnd_init(
-    bool scan
-  );
-#ifdef HA_CAN_BULK_ACCESS
-  int pre_rnd_init(
-    bool scan
-  );
-#endif
-  int rnd_end();
-#ifdef HA_CAN_BULK_ACCESS
-  int pre_rnd_end();
-#endif
+  int rnd_init(bool scan) override;
+  int rnd_end() override;
   int rnd_next(
     uchar *buf
-  );
+  ) override;
   void position(
     const uchar *record
-  );
+  ) override;
   int rnd_pos(
     uchar *buf,
     uchar *pos
-  );
+  ) override;
   int cmp_ref(
     const uchar *ref1,
     const uchar *ref2
-  );
-  int ft_init();
-  void ft_end();
+  ) override;
+  int ft_init() override;
+  void ft_end() override;
   FT_INFO *ft_init_ext(
     uint flags,
     uint inx,
     String *key
-  );
+  ) override;
   int ft_read(
     uchar *buf
-  );
+  ) override;
   int pre_index_read_map(
     const uchar *key,
     key_part_map keypart_map,
     enum ha_rkey_function find_flag,
     bool use_parallel
-  );
-  int pre_index_first(bool use_parallel);
-  int pre_index_last(bool use_parallel);
+  ) override;
+  int pre_index_first(bool use_parallel) override;
+  int pre_index_last(bool use_parallel) override;
   int pre_index_read_last_map(
     const uchar *key,
     key_part_map keypart_map,
     bool use_parallel
-  );
-#ifdef HA_MRR_USE_DEFAULT_IMPL
-  int pre_multi_range_read_next(
-    bool use_parallel
-  );
-#else
-  int pre_read_multi_range_first(
-    KEY_MULTI_RANGE **found_range_p,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    HANDLER_BUFFER *buffer,
-    bool use_parallel
-  );
-#endif
+  ) override;
+  int pre_multi_range_read_next(bool use_parallel) override;
   int pre_read_range_first(
     const key_range *start_key,
     const key_range *end_key,
     bool eq_range,
     bool sorted,
     bool use_parallel
-  );
-  int pre_ft_read(bool use_parallel);
-  int pre_rnd_next(bool use_parallel);
+  ) override;
+  int pre_ft_read(bool use_parallel) override;
+  int pre_rnd_next(bool use_parallel) override;
   int info(
     uint flag
-  );
+  ) override;
   ha_rows records_in_range(
     uint inx,
     const key_range *start_key,
     const key_range *end_key,
     page_range *pages
-  );
+  ) override;
   int check_crd();
-  int pre_records();
-  ha_rows records();
-#ifdef HA_HAS_CHECKSUM_EXTENDED
-  int pre_calculate_checksum();
-  int calculate_checksum();
-#endif
-  const char *table_type() const;
-  ulonglong table_flags() const;
+  int pre_records() override;
+  ha_rows records() override;
+  int pre_calculate_checksum() override;
+  int calculate_checksum() override;
+  const char *table_type() const override;
+  ulonglong table_flags() const override;
   ulong table_flags_for_partition();
-  const char *index_type(
-    uint key_number
-  );
-  ulong index_flags(
-    uint idx,
-    uint part,
-    bool all_parts
-  ) const;
-  uint max_supported_record_length() const;
-  uint max_supported_keys() const;
-  uint max_supported_key_parts() const;
-  uint max_supported_key_length() const;
-  uint max_supported_key_part_length() const;
-  uint8 table_cache_type();
-#ifdef HANDLER_HAS_NEED_INFO_FOR_AUTO_INC
-  bool need_info_for_auto_inc();
-#endif
+  const char *index_type(uint key_number) override;
+  ulong index_flags(uint idx, uint part, bool all_parts) const override;
+  uint max_supported_record_length() const override;
+  uint max_supported_keys() const override;
+  uint max_supported_key_parts() const override;
+  uint max_supported_key_length() const override;
+  uint max_supported_key_part_length() const override;
+  uint8 table_cache_type() override;
+  bool need_info_for_auto_inc() override;
 #ifdef HANDLER_HAS_CAN_USE_FOR_AUTO_INC_INIT
-  bool can_use_for_auto_inc_init();
+  bool can_use_for_auto_inc_init() override;
 #endif
   int update_auto_increment();
   void get_auto_increment(
@@ -498,51 +364,29 @@ public:
     ulonglong nb_desired_values,
     ulonglong *first_value,
     ulonglong *nb_reserved_values
-  );
-  int reset_auto_increment(
-    ulonglong value
-  );
-  void release_auto_increment();
-#ifdef SPIDER_HANDLER_START_BULK_INSERT_HAS_FLAGS
-  void start_bulk_insert(
-    ha_rows rows,
-    uint flags
-  );
-#else
-  void start_bulk_insert(
-    ha_rows rows
-  );
-#endif
-  int end_bulk_insert();
-  int write_row(
-    const uchar *buf
-  );
-#ifdef HA_CAN_BULK_ACCESS
-  int pre_write_row(
-    uchar *buf
-  );
-#endif
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+  ) override;
+  int reset_auto_increment(ulonglong value) override;
+  void release_auto_increment() override;
+  void start_bulk_insert(ha_rows rows, uint flags) override;
+  int end_bulk_insert() override;
+  int write_row(const uchar *buf) override;
   void direct_update_init(
     THD *thd,
     bool hs_request
   );
-#endif
-  bool start_bulk_update();
-  int exec_bulk_update(
-    ha_rows *dup_key_found
-  );
-  int end_bulk_update();
+  bool start_bulk_update() override;
+  int exec_bulk_update(ha_rows *dup_key_found) override;
+  int end_bulk_update() override;
 #ifdef SPIDER_UPDATE_ROW_HAS_CONST_NEW_DATA
   int bulk_update_row(
     const uchar *old_data,
     const uchar *new_data,
     ha_rows *dup_key_found
-  );
+  ) override;
   int update_row(
     const uchar *old_data,
     const uchar *new_data
-  );
+  ) override;
 #else
   int bulk_update_row(
     const uchar *old_data,
@@ -554,300 +398,107 @@ public:
     uchar *new_data
   );
 #endif
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
   bool check_direct_update_sql_part(
     st_select_lex *select_lex,
     longlong select_limit,
     longlong offset_limit
   );
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-#ifdef SPIDER_MDEV_16246
-  inline int direct_update_rows_init(
-    List<Item> *update_fields
-  ) {
-    return direct_update_rows_init(update_fields, 2, NULL, 0, FALSE, NULL);
-  }
-  int direct_update_rows_init(
-    List<Item> *update_fields,
-    uint mode,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    uchar *new_data
-  );
-#else
-  inline int direct_update_rows_init()
-  {
-    return direct_update_rows_init(2, NULL, 0, FALSE, NULL);
-  }
-  int direct_update_rows_init(
-    uint mode,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    uchar *new_data
-  );
-#endif
-#else
-#ifdef SPIDER_MDEV_16246
-  int direct_update_rows_init(
-    List<Item> *update_fields
-  );
-#else
-  int direct_update_rows_init();
-#endif
-#endif
-#ifdef HA_CAN_BULK_ACCESS
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-#ifdef SPIDER_MDEV_16246
-  inline int pre_direct_update_rows_init(
-    List<Item> *update_fields
-  ) {
-    return pre_direct_update_rows_init(update_fields, 2, NULL, 0, FALSE, NULL);
-  }
-  int pre_direct_update_rows_init(
-    List<Item> *update_fields,
-    uint mode,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    uchar *new_data
-  );
-#else
-  inline int pre_direct_update_rows_init()
-  {
-    return pre_direct_update_rows_init(2, NULL, 0, FALSE, NULL);
-  }
-  int pre_direct_update_rows_init(
-    uint mode,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    uchar *new_data
-  );
-#endif
-#else
-#ifdef SPIDER_MDEV_16246
-  int pre_direct_update_rows_init(
-    List<Item> *update_fields
-  );
-#else
-  int pre_direct_update_rows_init();
-#endif
-#endif
-#endif
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-  inline int direct_update_rows(ha_rows *update_rows, ha_rows *found_rows)
-  {
-    return direct_update_rows(NULL, 0, FALSE, NULL, update_rows, found_rows);
-  }
-  int direct_update_rows(
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    uchar *new_data,
-    ha_rows *update_rows,
-    ha_rows *found_row
-  );
-#else
-  int direct_update_rows(
-    ha_rows *update_rows,
-    ha_rows *found_row
-  );
-#endif
-#ifdef HA_CAN_BULK_ACCESS
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-  inline int pre_direct_update_rows()
-  {
-    ha_rows update_rows;
-    ha_rows found_rows;
-
-    return pre_direct_update_rows(NULL, 0, FALSE, NULL, &update_rows,
-      &found_rows);
-  }
-  int pre_direct_update_rows(
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    uchar *new_data,
-    ha_rows *update_rows,
-    ha_rows *found_row
-  );
-#else
-  int pre_direct_update_rows();
-#endif
-#endif
-#endif
-  bool start_bulk_delete();
-  int end_bulk_delete();
-  int delete_row(
-    const uchar *buf
-  );
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
+  int direct_update_rows_init(List<Item> *update_fields) override;
+  int direct_update_rows(ha_rows *update_rows, ha_rows *found_row) override;
+  bool start_bulk_delete() override;
+  int end_bulk_delete() override;
+  int delete_row(const uchar *buf) override;
   bool check_direct_delete_sql_part(
     st_select_lex *select_lex,
     longlong select_limit,
     longlong offset_limit
   );
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-  inline int direct_delete_rows_init()
-  {
-    return direct_delete_rows_init(2, NULL, 0, FALSE);
-  }
-  int direct_delete_rows_init(
-    uint mode,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted
-  );
-#else
-  int direct_delete_rows_init();
-#endif
-#ifdef HA_CAN_BULK_ACCESS
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-  inline int pre_direct_delete_rows_init()
-  {
-    return pre_direct_delete_rows_init(2, NULL, 0, FALSE);
-  }
-  int pre_direct_delete_rows_init(
-    uint mode,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted
-  );
-#else
-  int pre_direct_delete_rows_init();
-#endif
-#endif
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-  inline int direct_delete_rows(ha_rows *delete_rows)
-  {
-    return direct_delete_rows(NULL, 0, FALSE, delete_rows);
-  }
-  int direct_delete_rows(
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    ha_rows *delete_rows
-  );
-#else
-  int direct_delete_rows(
-    ha_rows *delete_rows
-  );
-#endif
-#ifdef HA_CAN_BULK_ACCESS
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS_WITH_HS
-  inline int pre_direct_delete_rows()
-  {
-    ha_rows delete_rows;
-
-    return pre_direct_delete_rows(NULL, 0, FALSE, &delete_rows);
-  }
-  int pre_direct_delete_rows(
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    ha_rows *delete_rows
-  );
-#else
-  int pre_direct_delete_rows();
-#endif
-#endif
-#endif
-  int delete_all_rows();
-  int truncate();
-  double scan_time();
+  int direct_delete_rows_init() override;
+  int direct_delete_rows(ha_rows *delete_rows) override;
+  int delete_all_rows() override;
+  int truncate() override;
+  double scan_time() override;
   double read_time(
     uint index,
     uint ranges,
     ha_rows rows
-  );
-#ifdef HA_CAN_BULK_ACCESS
-  void bulk_req_exec();
-#endif
-  const key_map *keys_to_use_for_scanning();
-  ha_rows estimate_rows_upper_bound();
+  ) override;
+  const key_map *keys_to_use_for_scanning() override;
+  ha_rows estimate_rows_upper_bound() override;
   void print_error(
     int error,
     myf errflag
-  );
+  ) override;
   bool get_error_message(
     int error,
     String *buf
-  );
+  ) override;
   int create(
     const char *name,
     TABLE *form,
     HA_CREATE_INFO *info
-  );
+  ) override;
   void update_create_info(
     HA_CREATE_INFO* create_info
-  );
+  ) override;
   int rename_table(
     const char *from,
     const char *to
-  );
+  ) override;
   int delete_table(
     const char *name
-  );
-  bool is_crashed() const;
+  ) override;
+  bool is_crashed() const override;
 #ifdef SPIDER_HANDLER_AUTO_REPAIR_HAS_ERROR
-  bool auto_repair(int error) const;
+  bool auto_repair(int error) const override;
 #else
   bool auto_repair() const;
 #endif
   int disable_indexes(
-    uint mode
-  );
+    key_map map, bool persist
+  ) override;
   int enable_indexes(
-    uint mode
-  );
+    key_map map, bool persist
+  ) override;
   int check(
     THD* thd,
     HA_CHECK_OPT* check_opt
-  );
+  ) override;
   int repair(
     THD* thd,
     HA_CHECK_OPT* check_opt
-  );
+  ) override;
   bool check_and_repair(
     THD *thd
-  );
+  ) override;
   int analyze(
     THD* thd,
     HA_CHECK_OPT* check_opt
-  );
+  ) override;
   int optimize(
     THD* thd,
     HA_CHECK_OPT* check_opt
-  );
+  ) override;
   bool is_fatal_error(
     int error_num,
     uint flags
-  );
+  ) override;
   Field *field_exchange(
     Field *field
   );
   const COND *cond_push(
     const COND* cond
-  );
-  void cond_pop();
-  int info_push(
-    uint info_type,
-    void *info
-  );
-#ifdef HANDLER_HAS_DIRECT_AGGREGATE
-  void return_record_by_parent();
-#endif
+  ) override;
+  void cond_pop() override;
+  int info_push(uint info_type, void *info) override;
+  void return_record_by_parent() override;
   TABLE *get_table();
   void set_ft_discard_bitmap();
   void set_searched_bitmap();
   void set_clone_searched_bitmap();
   void set_searched_bitmap_from_item_list();
   void set_select_column_mode();
-#ifdef WITH_PARTITION_STORAGE_ENGINE
   void check_select_column(bool rnd);
-#endif
   bool check_and_start_bulk_update(
     spider_bulk_upd_start bulk_upd_start
   );
@@ -860,25 +511,7 @@ public:
   bool is_sole_projection_field(
     uint16 field_index
   );
-  int check_ha_range_eof();
   int drop_tmp_tables();
-  bool handler_opened(
-    int link_idx,
-    uint tgt_conn_kind
-  );
-  void set_handler_opened(
-    int link_idx
-  );
-  void clear_handler_opened(
-    int link_idx,
-    uint tgt_conn_kind
-  );
-  int close_opened_handler(
-    int link_idx,
-    bool release_conn
-  );
-  int index_handler_init();
-  int rnd_handler_init();
   void set_error_mode();
   void backup_error_status();
   int check_error_mode(
@@ -907,34 +540,12 @@ public:
     bool eq_range,
     bool sorted
   );
-#ifdef HA_MRR_USE_DEFAULT_IMPL
-#else
-  int read_multi_range_first_internal(
-    uchar *buf,
-    KEY_MULTI_RANGE **found_range_p,
-    KEY_MULTI_RANGE *ranges,
-    uint range_count,
-    bool sorted,
-    HANDLER_BUFFER *buffer
-  );
-#endif
   int ft_read_internal(uchar *buf);
   int rnd_next_internal(uchar *buf);
   void check_pre_call(
     bool use_parallel
   );
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
   void check_insert_dup_update_pushdown();
-#endif
-#ifdef HA_CAN_BULK_ACCESS
-  SPIDER_BULK_ACCESS_LINK *create_bulk_access_link();
-  void delete_bulk_access_link(
-    SPIDER_BULK_ACCESS_LINK *bulk_access_link
-  );
-  int sync_from_clone_source(
-    ha_spider *spider
-  );
-#endif
   void sync_from_clone_source_base(
     ha_spider *spider
   );
@@ -954,10 +565,7 @@ public:
   int append_insert_sql_part();
   int append_update_sql_part();
   int append_update_set_sql_part();
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
   int append_direct_update_set_sql_part();
-#endif
-#ifdef HANDLER_HAS_DIRECT_UPDATE_ROWS
   int append_dup_update_pushdown_sql_part(
     const char *alias,
     uint alias_length
@@ -967,7 +575,6 @@ public:
     uint alias_length
   );
   int check_update_columns_sql_part();
-#endif
   int append_delete_sql_part();
   int append_select_sql_part(
     ulong sql_type
@@ -1031,13 +638,11 @@ public:
     ulong sql_type,
     bool test_flg
   );
-#ifdef HANDLER_HAS_DIRECT_AGGREGATE
   int append_sum_select_sql_part(
     ulong sql_type,
     const char *alias,
     uint alias_length
   );
-#endif
   int append_match_select_sql_part(
     ulong sql_type,
     const char *alias,
@@ -1049,13 +654,11 @@ public:
   void set_order_to_pos_sql(
     ulong sql_type
   );
-#ifdef HANDLER_HAS_DIRECT_AGGREGATE
   int append_group_by_sql_part(
     const char *alias,
     uint alias_length,
     ulong sql_type
   );
-#endif
   int append_key_order_for_merge_with_alias_sql_part(
     const char *alias,
     uint alias_length,
@@ -1152,21 +755,32 @@ public:
     const char *alias,
     uint alias_length
   );
-  bool support_use_handler_sql(
-    int use_handler
-  );
   int init_union_table_name_pos_sql();
   int set_union_table_name_pos_sql();
   int append_lock_tables_list();
   int lock_tables();
   int dml_init();
-#ifdef HA_CAN_BULK_ACCESS
-  int bulk_access_begin(
-    void *info
-  );
-  int bulk_access_current(
-    void *info
-  );
-  void bulk_access_end();
-#endif
 };
+
+
+/* This is a hack for ASAN
+ * Libraries such as libxml2 and libodbc do not like being unloaded before
+ * exit and will show as a leak in ASAN with no stack trace (as the plugin
+ * has been unloaded from memory).
+ *
+ * The below is designed to trick the compiler into adding a "UNIQUE" symbol
+ * which can be seen using:
+ * readelf -s storage/spider/ha_spider.so | grep UNIQUE
+ *
+ * Having this symbol means that the plugin remains in memory after dlclose()
+ * has been called. Thereby letting the libraries clean up properly.
+ */
+#if defined(__SANITIZE_ADDRESS__)
+__attribute__((__used__))
+inline int dummy(void)
+{
+  static int d;
+  d++;
+  return d;
+}
+#endif
